/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <cmath>
#include <complex>
#include <assert.h>

// testing
#include <QApplication>
#include <QString>

#include "signalprocessing.h"

static const double						ZERO_EPSILON = 0.000001;

TFilterBase::TFilterBase(void)
{
}

TFilterBase::~TFilterBase(void)
{
}

double TFilterBase::Calculate_Weight(
	const double						&x,
	const double						&cutoff_length) const
{
	static const double					filter_constant(0.46971883772713204146);	//sqrt(ln(2)/3.14159)
	static const double					pi(3.14159265358979323846);
	double								exponent;
	double								denom;
	
	assert(cutoff_length > ZERO_EPSILON);
	
	exponent = x/(filter_constant * cutoff_length);
	exponent *= exponent;
	exponent *= pi;
	
	denom = filter_constant * cutoff_length;
	
	return ((1.0 / denom) * exp(-exponent));
}


std::vector<TVector2> TFilterBase::Filter_Data(
	const std::vector<TVector2>			&data,
	const double						&cutoff_length) const
{
	std::vector<TVector2>				filtered_data;
	std::vector<TVector2>::const_iterator iter,sub_iter;
	int									index;
	int									offset;
	double								weight;
	double								weight_zero;
	double								sum_weight;
	double								sum_data;
	TVector2							zero_data;
	TVector2							item_data;
	TVector2							item_data_end;
	double								weight_low_limit(0.01);	// limit when value of weight becomes insignificant.  0.010 = 1/100 th of current value
	
	assert(cutoff_length > ZERO_EPSILON);
	
	weight_zero = this->Calculate_Weight(0.0,cutoff_length);
	
	for(iter = data.begin();iter != data.end();++iter)
	{
		sum_weight = 0;
		sum_data = 0;
		
		zero_data = (*iter);
		
		sum_weight += 1;
		sum_data += (zero_data.y);
		
		// calculate negative side weights
		// use reflection to fill in missing parts
		
		index = 1;
		do{
			
			sub_iter = (iter - index);
			
			if(sub_iter < data.begin())
			{
				offset = (data.begin() - sub_iter);	// position where data is used for reflection
				
				item_data = (*(data.begin() + offset));
				item_data_end = (*(data.begin()));
				
				item_data -= item_data_end;
				item_data *= -1;
				item_data += item_data_end;
			}
			else
			{
				item_data = (*sub_iter);
			}
			
			++index;
			
			weight = this->Calculate_Weight(item_data.x - zero_data.x,cutoff_length) / weight_zero;
			
			sum_data += (item_data.y * weight);
			sum_weight += weight;
			
			
		}while (!(weight < weight_low_limit));
		
		// calculate positive side weights
		// use reflection to fill in missing parts
		
		index = 1;
		do{
			
			sub_iter = (iter + index);
			
			if(!(sub_iter < data.end()))
			{
				offset = (sub_iter - data.end());	// position where data is used for reflection
				
				item_data = (*(data.end() - offset - 2));
				item_data_end = (*(data.end() - 1));
				
				item_data -= item_data_end;
				item_data *= -1;
				item_data += item_data_end;
			}
			else
			{
				item_data = (*sub_iter);
			}
			
			++index;
			
			weight = this->Calculate_Weight(item_data.x - zero_data.x,cutoff_length) / weight_zero;

			sum_data += (item_data.y * weight);
			sum_weight += weight;
			
		}while (!(weight < weight_low_limit));
		
		sum_data /= sum_weight;
		
		zero_data.y = sum_data;
		
		filtered_data.push_back(zero_data);
	}
	
	return filtered_data;
}


TLowPassFilter::TLowPassFilter(void)
:TFilterBase()
{
	d_cutoff_length = 1;
}

TLowPassFilter::~TLowPassFilter(void)
{
}

std::vector<TVector2> TLowPassFilter::Filter_Data_Set(
	const std::vector<TVector2>			&data) const
{
	return this->Filter_Data(data,d_cutoff_length);
}


THighPassFilter::THighPassFilter(void)
:TFilterBase()
{
	d_cutoff_length = 1;
}

THighPassFilter::~THighPassFilter(void)
{
}

std::vector<TVector2> THighPassFilter::Filter_Data_Set(
	const std::vector<TVector2>			&data) const
{
	std::vector<TVector2>				filtered_data;
	std::vector<TVector2>::const_iterator data_iter;
	std::vector<TVector2>::iterator		filtered_iter;
	
	filtered_data = this->Filter_Data(data,d_cutoff_length);
	
	assert(filtered_data.size() == data.size());
	
	// high pass is identical to low pass except for the part of the data that is kept
	for(data_iter = data.begin(),filtered_iter = filtered_data.begin();data_iter != data.end();++data_iter,++filtered_iter)
	{
		(*filtered_iter).y = (*data_iter).y - (*filtered_iter).y;
	}
	
	return filtered_data;
}


TBandPassFilter::TBandPassFilter(void)
:TFilterBase()
{
	d_cutoff_length_low = 1;
	d_cutoff_length_high = 1;
}

TBandPassFilter::~TBandPassFilter(void)
{
}

std::vector<TVector2> TBandPassFilter::Filter_Data_Set(
	const std::vector<TVector2>			&data) const
{
	std::vector<TVector2>				filtered_data_low;
	std::vector<TVector2>				filtered_data_high;
	std::vector<TVector2>::const_iterator data_iter;
	std::vector<TVector2>::iterator		filtered_low_iter;
	std::vector<TVector2>::iterator		filtered_high_iter;
	
	filtered_data_low = this->Filter_Data(data,d_cutoff_length_low);
	filtered_data_high = this->Filter_Data(data,d_cutoff_length_high);
	
	assert(filtered_data_low.size() == data.size());
	assert(filtered_data_high.size() == data.size());

	// get the high pass data
	for(data_iter = data.begin(),filtered_high_iter = filtered_data_high.begin();data_iter != data.end();++data_iter,++filtered_high_iter)
	{
		(*filtered_high_iter).y = (*data_iter).y - (*filtered_high_iter).y;
	}
	
	// combine the high and low poass
	for(filtered_low_iter = filtered_data_low.begin(),filtered_high_iter = filtered_data_high.begin();
		filtered_low_iter != filtered_data_low.end();
		++filtered_low_iter,++filtered_high_iter)
	{
		(*filtered_low_iter).y += (*filtered_high_iter).y;
	}

	
	// subract the inverted high and low filters from the original data
	for(data_iter = data.begin(),filtered_low_iter = filtered_data_low.begin();
		data_iter != data.end();
		++data_iter,++filtered_low_iter)
	{
		(*filtered_low_iter).y -= (*data_iter).y;
	}
	
	return filtered_data_low;
}

void TBandPassFilter::Configure_Filter(
	const double						&filter_cutoff_length_low,
	const double						&filter_cutoff_length_high)
{
	if(filter_cutoff_length_high > filter_cutoff_length_low)
	{
		d_cutoff_length_high = filter_cutoff_length_high;
		d_cutoff_length_low = filter_cutoff_length_low;
	}
	else	// inputs are reversed
	{
		d_cutoff_length_high = filter_cutoff_length_low;
		d_cutoff_length_low = filter_cutoff_length_high;
	}
}

TKalmanFilter::TKalmanFilter(void)
{
	d_q = 0.000001;
	d_r = 0.01;
	d_p = 1.0;
	d_x = 0;
	d_k = 0;
}

TKalmanFilter::~TKalmanFilter(void)
{
}

double TKalmanFilter::Stream_In(
	const double						&data)
{
	double								result;
	
	this->Update();
	
	result = d_x + (data - d_x) * d_k;
	d_x = result;
	
	return result;
}

void TKalmanFilter::Update(void)
{
	d_k = (d_p + d_q) / (d_p + d_q + d_r);
	d_p = d_r * (d_p + d_q) / (d_r + d_p + d_q);
}

TIntegrator::TIntegrator(void)
{
	d_init = 0;
}

TIntegrator::~TIntegrator(void)
{
}

void TIntegrator::Add_Data(
	const double						&data,
	const double						&data_increment)
{
	double								area;
	const double						ZERO_EPSILON(0.000001);
	
	// calculate area using a trapezoid
	// area is dx * (a+b)/2
	
	if(d_init)
	{
		d_last_input_data = data;
		d_sum = 0;
		
		d_init = false;
	}
	else
	{
		if(data_increment > ZERO_EPSILON)
		{
			area = data_increment * (d_last_input_data + data) / 2.0;
			d_sum += area;
			
			d_last_input_data = data;
		}
	}
}

/*
 based on the Cooley-Tukey recursive FFT algorithm posted on Wikipedia
*/
TFourierTransform::TFourierTransform(void)
{
}

TFourierTransform::~TFourierTransform(void)
{
}

std::vector<TVector2> TFourierTransform::Process(
	const std::vector<TVector2>			&in_data)
{
	double								dval;
	double								sample_rate;
	double								sample_increment;
	double								frequency_step;
	double								time_start(0);
	double								time_increment(0);
	int									exponent(0);
	unsigned int						in_data_size;
	unsigned int						out_data_size;
	unsigned int						out_data_return_size;
	unsigned int						cntr;
	std::vector<TVector2>				out_data;
	std::complex<double>				*result;
	static const double					LESS_THAN_ONE(1.0 - ZERO_EPSILON);
	
	in_data_size = in_data.size();
	
	if(in_data_size < 2)
	{
		return out_data;
	}
	
	// time information
	time_start = (*in_data.begin()).x;
	time_increment = ((*(in_data.end() - 1)).x - time_start);

	// calculate necessary data size.  must be a power of two
	dval = log(static_cast<double>(in_data_size)) / log(2.0);
	exponent = static_cast<int>(dval + LESS_THAN_ONE);
	
	out_data_size = static_cast<unsigned int>(pow(2.0,exponent) + 0.5);
	out_data.resize(out_data_size);
	out_data_return_size = out_data_size / 2;
	
	if(time_increment > ZERO_EPSILON)
	{
		sample_rate = static_cast<double>(in_data_size - 1) / time_increment;
		sample_increment = sample_rate / static_cast<double>(out_data_size);// * static_cast<double>(in_data_size) / static_cast<double>(out_data_size);
		frequency_step = 1.0 / sample_increment;
		
//		qWarning(QString("sample_rate: %1  time_increment: %2  in_data_size: %3").arg(sample_rate,0,'f',4).arg(time_increment).arg(in_data_size).toLatin1());
	}
	else
	{
		return out_data;
	}
	
	result = new std::complex<double>[out_data_size];
	
	for(cntr = 0;cntr < in_data_size;++cntr)
	{
		result[cntr] = std::complex<double>(in_data[cntr].y,0.0);
	}
	
	for(cntr = in_data_size;cntr < out_data_size;++cntr)
	{
		result[cntr] = std::complex<double>(0.,0.);
	}
	
	this->Transform(result,out_data_size);
	
	
	// copy results to transformed array
	for(cntr = 0;cntr < out_data_size;++cntr)
	{
		out_data[cntr] = TVector2(result[cntr].real(),abs(result[cntr]) * sample_increment);
	}
	
	delete[] result;
	
	out_data.resize(out_data_return_size);

	// update frequency component
	for(cntr = 0;cntr < out_data_return_size;++cntr)
	{
		out_data[cntr].x = static_cast<double>(cntr) / frequency_step;
	}

	
	return out_data;
}

void TFourierTransform::Separate(
	std::complex<double>				* const data,
	const int							size)
{
	int									half_size(size / 2);
	int									cntr;
	std::complex<double>				*temp_data;
	
	temp_data = new std::complex<double>[half_size];
	
	// move all odd elements to lower half
	// move all even elements to upper half
	
	for(cntr=0; cntr < half_size;++cntr)
	{
		temp_data[cntr] = data[cntr * 2 + 1];	// odd
	}
	
	for(cntr=0; cntr < half_size;++cntr)
	{
		data[cntr] = data[cntr * 2];			// even
	}
	
	for(cntr=0; cntr < half_size;++cntr)
	{
		data[cntr + half_size] = temp_data[cntr];
	}
	
	delete[] temp_data;
}

void TFourierTransform::Transform(
	std::complex<double>				* const data,
	const int							size)
{
	int									half_size(size / 2);
	int									cntr;
	static const double					PI(3.141592653589793);
	
	if(!(size < 2))
	{
		this->Separate(data,size);						// all evens to lower half, all odds to upper half
		
		this->Transform(data,half_size);				// recurse even items
		this->Transform(data + half_size, half_size);	// recurse odd  items
		
		// combine results of two half recursions
		for(cntr = 0; cntr < half_size;++cntr)
		{
			std::complex<double> e = data[cntr];   // even
			std::complex<double> o = data[cntr + half_size];   // odd
			
			// w is the "twiddle-factor"
			std::complex<double> w = exp(std::complex<double>(0.0,-2.0 * PI * static_cast<double>(cntr) / static_cast<double>(size)));
		
			data[cntr] = e + w * o;
			data[cntr + half_size] = e - w * o;
		}
	}
}

