/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef SIGNALPROCESSINGHEADERFILE
#define SIGNALPROCESSINGHEADERFILE

#include <vector>
#include <complex>

#include "vector2.h"

class TFilterBase
{
public:
// CREATORS
	TFilterBase(void);
	virtual ~TFilterBase(void)=0;
	
// ACCESSORS
	
// MANIPULATORS

protected:
	
	double Calculate_Weight(const double &x,const double &cutoff_length) const;
	std::vector<TVector2> Filter_Data(const std::vector<TVector2> &data, const double &cutoff_length) const;

// NOT IMPLEMENTED
};

class TLowPassFilter : public TFilterBase
{
public:
	
// CREATORS
	TLowPassFilter(void);
	virtual ~TLowPassFilter(void);
	
// ACCESSORS
	double Cutoff_Length(void) const {return d_cutoff_length;}
	std::vector<TVector2> Filter_Data_Set(const std::vector<TVector2> &data) const;
	
// MANIPULATORS
	void Configure_Filter(const double &filter_cutoff_length) {d_cutoff_length = filter_cutoff_length;}
	
private:
	double								d_cutoff_length;
	
// NOT IMPLEMENTED
	
};

class THighPassFilter : public TFilterBase
{
public:
	
	// CREATORS
	THighPassFilter(void);
	virtual ~THighPassFilter(void);
	
	// ACCESSORS
	double Cutoff_Length(void) const {return d_cutoff_length;}
	std::vector<TVector2> Filter_Data_Set(const std::vector<TVector2> &data) const;
	
	// MANIPULATORS
	void Configure_Filter(const double &filter_cutoff_length) {d_cutoff_length = filter_cutoff_length;}
	
private:
	double								d_cutoff_length;
	
	// NOT IMPLEMENTED
};

class TBandPassFilter : public TFilterBase
{
public:
	
	// CREATORS
	TBandPassFilter(void);
	virtual ~TBandPassFilter(void);
	
	// ACCESSORS
	double Cutoff_Length_Low(void) const {return d_cutoff_length_low;}
	double Cutoff_Length_High(void) const {return d_cutoff_length_high;}
	std::vector<TVector2> Filter_Data_Set(const std::vector<TVector2> &data) const;
	
	// MANIPULATORS
	void Configure_Filter(const double &filter_cutoff_length_low,const double &filter_cutoff_length_high);
	
private:
	double								d_cutoff_length_low;
	double								d_cutoff_length_high;
	
	// NOT IMPLEMENTED
};

class TKalmanFilter
{
public:
	
// CREATORS
	TKalmanFilter(void);
	~TKalmanFilter(void);
	
// ACCESSORS
	
// MANIPULATORS
	double Stream_In(const double &data);
	
private:
	double								d_q;
	double								d_r;
	double								d_p;
	double								d_x;
	double								d_k;
	
	void Update(void);
// NOT IMPLEMENTED
};


class TIntegrator
{
public:
	
// CREATORS
	TIntegrator(void);
	~TIntegrator(void);
	
// ACCESSORS
	double Result(void) const {return d_sum;}
	
// MANIPULATORS
	void Reset(void) {d_init = true;}
	
	void Add_Data(const double &data,const double &data_increment);		// data_increment is difference from last input data
	
private:
	
	double								d_sum;
	double								d_last_input_data;
	bool								d_init;
	
// NOT IMPLEMENTED
};

class TFourierTransform
{
public:
	
// CREATORS
	TFourierTransform(void);
	~TFourierTransform(void);
	
// ACCESSORS

// MANIPULATORS
	
	// x value is sample time in seconds, y value is amplitude.
	// return x value is frequency, y value is amplitude
	std::vector<TVector2> Process(const std::vector<TVector2> &data);
	
private:
	
	void Separate(std::complex<double> * const data,const int size);
	void Transform(std::complex<double>	* const data,const int	size);

// NOT IMPLEMENTED
};

#endif
